# -*- coding: iso-8859-1 -*-
"""
    MoinMoin - MoinMessage library

    @copyright: 2012, 2013, 2014, 2015 by Paul Boddie <paul@boddie.org.uk>
    @license: GNU GPL (v2 or later), see COPYING.txt for details.
"""

from email import message_from_string
from email.encoders import encode_noop
from email.generator import Generator
from email.mime.multipart import MIMEMultipart
from email.mime.application import MIMEApplication
from email.mime.base import MIMEBase
from email.mime.text import MIMEText
from email.parser import Parser
from email.utils import formatdate
from itertools import islice
from subprocess import Popen, PIPE
from tempfile import mkstemp
from urlparse import urlsplit
from GPGUtils import GPG, GPGError, GPGDecodingError, GPGMissingPart, GPGBadContent, \
                     as_string, is_signed, is_encrypted, getContentAndSignature
import httplib
import os

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

# Message inspection functions.

def is_collection(message):
    return message.get("Update-Type") == "collection"

def to_replace(message):
    return message.get("Update-Action") == "replace"

def to_store(message):
    return message.get("Update-Action") == "store"

def get_update_action(message):
    return message.get("Update-Action", "update")

# Core abstractions.

class Message:

    "An update message."

    def __init__(self, text=None):
        self.date = None
        self.updates = []
        if text:
            self.parse_text(text)

    def init_date(self, message):

        "Obtain the date of the given 'message'."

        self.date = message.get("Date")

    def parse_text(self, text):

        "Parse the given 'text' as a message."

        self.handle_message(message_from_string(text))

    def handle_message(self, message):

        "Handle the given 'message', recording the separate updates."

        self.init_date(message)

        # The message either consists of a collection of updates.

        if message.is_multipart() and is_collection(message):
            for part in message.get_payload():
                self.updates.append(part)

        # Or the message is a single update.

        else:
            self.updates.append(message)

    def add_updates(self, parts):

        """
        Add the given 'parts' to a message.
        """

        for part in updates:
            self.add_update(part)

    def add_update(self, part):

        """
        Add an update 'part' to a message.
        """

        self.updates.append(part)

    def get_update(self, alternatives):

        """
        Return a suitable multipart object containing the supplied
        'alternatives'.
        """

        part = MIMEMultipart("alternative")
        for alternative in alternatives:
            part.attach(alternative)
        return part

    def get_payload(self, subtype="mixed", timestamped=True):

        """
        Get the multipart payload for the message. If the 'timestamped'
        parameter is omitted or set to a true value, the payload will be given a
        date header set to the current date and time that can be used to assess
        the validity of a message and to determine whether it has already been
        received by a recipient.
        """

        if len(self.updates) == 1:
            message = self.updates[0]
        else:
            message = MIMEMultipart(subtype)
            message.add_header("Update-Type", "collection")
            for update in self.updates:
                message.attach(update)

        if timestamped:
            timestamp(message)
            self.init_date(message)

        return message

MoinMessageError = GPGError
MoinMessageDecodingError = GPGDecodingError
MoinMessageMissingPart = GPGMissingPart
MoinMessageBadContent = GPGBadContent

class MoinMessageTransferError(MoinMessageError):
    def __init__(self, code, message, body):
        MoinMessageError.__init__(self, message)
        self.code = code
        self.body = body

# Communications functions.

def timestamp(message):

    """
    Timestamp the given 'message' so that its validity can be assessed by the
    recipient.
    """

    datestr = formatdate()

    if not message.has_key("Date"):
        message.add_header("Date", datestr)
    else:
        message.replace_header("Date", datestr)

def _getConnection(scheme):

    "Return the connection class for the given 'scheme'."

    if scheme == "http":
        return httplib.HTTPConnection
    elif scheme == "https":
        return httplib.HTTPSConnection
    else:
        raise MoinMessageError, "Communications protocol not supported: %s" % scheme

def sendMessageOpener(message, url, method="PUT"):

    """
    Send 'message' to the given 'url' using the given 'method' (using PUT as the
    default if omitted).
    """

    scheme, host, port, path = parseURL(url)
    text = as_string(message)

    req = _getConnection(scheme)(host, port)
    req.request(method, path, text)
    resp = req.getresponse()

    if resp.status >= 400:
        raise MoinMessageTransferError(resp.status, "Message sending failed (%s)" % resp.status, resp.read())

    return resp

def sendMessage(message, url, method="PUT"):

    """
    Send 'message' to the given 'url' using the given 'method' (using PUT as the
    default if omitted).
    """

    resp = sendMessageOpener(message, url, method)
    return resp.read()

def parseURL(url):

    "Return the scheme, host, port and path for the given 'url'."

    scheme, host_port, path, query, fragment = urlsplit(url)
    host_port = host_port.split(":")

    if query:
        path += "?" + query

    if len(host_port) > 1:
        host = host_port[0]
        port = int(host_port[1])
    else:
        host = host_port[0]
        port = 80

    return scheme, host, port, path

# Message handling.

class MessageInterface:

    "A command-based interface to a message store, inspired by RFC 1939 (POP3)."

    def __init__(self, store):
        self.store = store

    def execute(self, commands):

        """
        Access messages according to the given 'commands' script, acting on the
        store provided during initialisation and returning a message object
        containing the results.
        """

        # Build a container for the responses.

        message = Message()

        # Process each command.

        for command in commands.split("\n"):
            command = command.strip()

            # Get the command and arguments.

            command_parts = command.split(None, 1)
            cmd = command_parts[0]

            try:
                if cmd in self.commands:
                    getattr(self, cmd)(command_parts, message)
                else:
                    self.add_result(cmd, command, "ERR", message)
            except Exception, exc:
                self.add_result(cmd, "\n".join([command, str(exc)]), "ERR", message)

        return message

    def get_count(self, command_parts):

        # Select all messages by default.

        count = None

        if len(command_parts) > 1:
            count = int(command_parts[1])

        return count

    def add_result(self, cmd, result, status, message):
        part = MIMEText(result, "x-moinmessage-fetch-result")
        part["Request-Type"] = cmd
        part["Request-Status"] = status
        message.add_update(part)

    def add_messages(self, resources, message):
        container = Message()

        for message_text in resources:
            message_item = Parser().parsestr(message_text)
            container.add_update(message_item)

        # Convert the container to a proper multipart section.

        message.add_update(container.get_payload())

    def STAT(self, command_parts, message):

        # A request to count the messages is returned in a part.

        self.add_result("STAT", str(len(self.store)), "OK", message)

    def RETR(self, command_parts, message):

        # A request for specific messages returns each message
        # in its own part within a collection part.

        count = self.get_count(command_parts)

        self.add_messages(islice(iter(self.store), count), message)

    def DELE(self, command_parts, message):

        # A request to delete messages is performed immediately.

        count = self.get_count(command_parts)

        keys = self.store.keys()[:count]

        for key in keys:
            del self.store[key]

        self.add_result("DELE", str(len(keys)), "OK", message)

    # Command manifest, may be extended by subclasses.

    commands = "STAT", "RETR", "DELE"

# vim: tabstop=4 expandtab shiftwidth=4
