# -*- coding: iso-8859-1 -*-
"""
    MoinMoin - MoinMessageSupport library

    @copyright: 2012, 2013, 2014, 2015 by Paul Boddie <paul@boddie.org.uk>
    @license: GNU GPL (v2 or later), see COPYING.txt for details.
"""

from MoinMoin import config, user
from MoinMoin.Page import Page
from MoinMoin.action import AttachFile
from MoinMoin.formatter import text_html
from MoinMoin.log import getLogger
from MoinMoin.user import User
from MoinMoin.wikiutil import parseQueryString, taintfilename, \
                              version2timestamp, getInterwikiHomePage

from MoinMessage import GPG, Message, MoinMessageError, \
                        MoinMessageMissingPart, MoinMessageBadContent, \
                        is_signed, is_encrypted, getContentAndSignature
from MoinSupport import getHeader, getMetadata, getWikiDict, writeHeaders, \
                        parseDictEntry, getStaticContentDirectory
from DateSupport import getDateTimeFromRFC2822
from ItemSupport import ItemStore
from TokenSupport import getIdentifiers

from email.parser import Parser
from os.path import abspath, exists, join
import time
import urllib

RECIPIENT_PARAMETERS = ("type", "location", "fingerprint")

class MoinMessageAction:

    "Common message handling support for actions."

    def __init__(self, pagename, request):

        """
        On the page with the given 'pagename', use the given 'request' when
        reading posted messages, modifying the Wiki.
        """

        self.pagename = pagename
        self.request = request
        self.page = Page(request, pagename)
        self.message = None
        self.new_user = None

    def init_store(self):
        self.store = ItemStore(self.page, "messages", "message-locks")

    def do_as_user(self, new_user, fn, args):

        "As 'new_user', perform 'fn' using the given 'args'."

        request = self.request

        # Switch to the relaying user if necessary.

        if new_user:
            user = request.user
            request.user = get_user(request, new_user) or user

        # Handle the content.

        try:
            fn(*args)
        finally:
            if new_user:
                request.user = user

    def do_action(self):
        request = self.request
        content_length = getHeader(request, "Content-Length", "HTTP")
        if content_length:
            content_length = int(content_length)

        self.handle_message_text(request.read(content_length))

    def handle_message_text(self, message_text):

        "Handle the given 'message_text'."

        request = self.request
        message = Parser().parsestr(message_text)

        # Detect any indicated recipient and change the target page, if
        # appropriate.

        new_user = None

        if message.has_key("To"):
            try:
                parameters = get_recipient_details(request, message["To"], main=True)
            except MoinMessageRecipientError, exc:

                # Reject missing recipients if being strict and not relying only
                # on signatures and user actions.

                if getattr(request, "moinmessage_reject_missing_global_recipients", False):
                    writeHeaders(request, "text/plain", getMetadata(self.page), "403 Forbidden")
                    request.write("The recipient indicated in the message is not known to this site. "
                                  "Details: %s" % exc.message)
                    return
            else:
                # Recipients with pages can have their messages redirected to
                # those pages.

                if parameters["type"] == "page":
                    self.page = Page(request, parameters["location"])

                # Recipients with URLs can have their messages forwarded.

                # Recipients accessible via relays have their messages
                # forwarded.

                elif parameters["type"] in ("url", "relay"):

                    # Get the relaying user page and select it.

                    relaying_user = get_relaying_user(request)
                    relaying_page = relaying_user and get_local_homepage(request, relaying_user)

                    if not relaying_page:
                        writeHeaders(request, "text/plain", getMetadata(self.page), "403 Forbidden")
                        request.write("This site is not able to forward the message to the recipient.")
                        return

                    # Stored messages should then be processed asynchronously.

                    self.page = Page(request, relaying_page)
                    new_user = relaying_user

        # Remember the original message for later processing.

        self.message = message
        self.new_user = new_user

        # Handle the parsed message.

        self.init_store()
        self.handle_message(message)

    def handle_message(self, message):

        "Handle the given 'message'."

        # Detect PGP/GPG-encoded payloads.
        # See: http://tools.ietf.org/html/rfc3156

        # Signed payloads are checked and then passed on for further processing
        # elsewhere. Verification is the last step in this base implementation,
        # even if an encrypted-then-signed payload is involved.

        if is_signed(message):
            self.handle_signed_message(message)

        # Encrypted payloads are decrypted and then sent back into this method
        # for signature checking as described above. Thus, signed-then-encrypted
        # payloads are first decrypted and then verified.

        elif is_encrypted(message):
            self.handle_encrypted_message(message)

        # Reject unsigned and unencrypted payloads.

        else:
            request = self.request
            writeHeaders(request, "text/plain", getMetadata(self.page), "415 Unsupported Media Type")
            request.write("Only PGP/GPG-signed payloads are supported.")

    def handle_encrypted_message(self, message):

        "Handle the given encrypted 'message'."

        request = self.request

        homedir = self.get_homedir()
        if not homedir:
            writeHeaders(request, "text/plain", getMetadata(self.page), "403 Forbidden")
            request.write("This site is not configured for this request.")
            return

        gpg = GPG(homedir)

        try:
            text = gpg.decryptMessage(message)

        # Reject messages without a declaration.

        except MoinMessageMissingPart:
            writeHeaders(request, "text/plain", getMetadata(self.page), "415 Unsupported Media Type")
            request.write("There must be a declaration and a content part for encrypted uploads.")
            return

        # Reject messages without appropriate content.

        except MoinMessageBadContent:
            writeHeaders(request, "text/plain", getMetadata(self.page), "415 Unsupported Media Type")
            request.write("Encrypted data must be provided as application/octet-stream.")
            return

        # Reject any undecryptable message.

        except MoinMessageError:
            writeHeaders(request, "text/plain", getMetadata(self.page), "403 Forbidden")
            request.write("The message could not be decrypted.")
            return

        # Log non-fatal errors.

        if gpg.errors:
            getLogger(__name__).warning(gpg.errors)

        # Handle the embedded message which may itself be a signed message.

        self.handle_message_text(text)

    def handle_signed_message(self, message):

        "Handle the given signed 'message'."

        request = self.request

        # Accept any message whose sender was authenticated by the PGP method.

        if request.user and request.user.valid and request.user.auth_method == "pgp":

            # Handle the embedded message.

            content, signature = getContentAndSignature(message)
            self.handle_message_content(content)

        # Reject any unverified message.

        else:
            writeHeaders(request, "text/plain", getMetadata(self.page), "403 Forbidden")
            request.write("The message could not be verified. "
                          "Maybe this site is not performing authentication using PGP signatures.")

    def handle_message_content(self, content):

        "Handle the given message 'content'."

        request = self.request

        # Interpret the content as one or more updates.

        message = Message()
        message.handle_message(content)

        # Test any date against the page or message store (if not empty).

        if message.date:
            date = getDateTimeFromRFC2822(message.date)
            store_date = time.gmtime(self.store.mtime())
            page_date = time.gmtime(version2timestamp(self.page.mtime_usecs()))

            if len(self.store) > 0:
                last_date = max(store_date, page_date)
            else:
                last_date = page_date

            # Reject messages older than the page date.

            if date.to_utc().as_tuple() < last_date:
                writeHeaders(request, "text/plain", getMetadata(self.page), "403 Forbidden")
                request.write("The message is too old: %s versus %s." % (date, time.strftime("%Y-%m-%d %H:%M:%S", last_date)))
                return

        # Reject messages without dates if so configured.

        elif getattr(request.cfg, "moinmessage_reject_messages_without_dates", True):
            writeHeaders(request, "text/plain", getMetadata(self.page), "403 Forbidden")
            request.write("The message has no date information.")
            return

        # Handle the message as an object.

        self.handle_message_object(message)

    def get_homedir(self):

        "Locate the GPG home directory."

        request = self.request
        homedir = get_homedir(self.request)

        if not homedir:
            writeHeaders(request, "text/plain", getMetadata(self.page), "415 Unsupported Media Type")
            request.write("Encoded data cannot currently be understood. Please notify the site administrator.")

        return homedir

    def can_perform_action(self, action):

        """
        Determine whether the user in the request has the necessary privileges
        to change the current page using a message requesting the given
        'action'.
        """

        for identifier in get_update_actions_for_user(self.request):

            # Expect "action:pagename", rejecting ill-formed identifiers.

            details = identifier.split(":", 1)
            if len(details) != 2:
                continue

            # If the action and page name match, return success.

            permitted, pagename = details
            if permitted.lower() == action.lower() and pagename == self.page.page_name:
                return True

        return False

# More specific errors.

class MoinMessageRecipientError(MoinMessageError):
    pass

class MoinMessageNoRecipients(MoinMessageRecipientError):
    pass

class MoinMessageUnknownRecipient(MoinMessageRecipientError):
    pass

class MoinMessageBadRecipient(MoinMessageRecipientError):
    pass

# Utility functions.

def get_user(request, username):

    "Return the user having the given 'username'."

    uid = user.getUserId(request, username)

    # If the user does not exist, just return None.

    if not uid:
        return None

    # Otherwise, return the requested user.

    return user.User(request, uid)

def get_homedir(request):

    "Locate the GPG home directory."

    return getattr(request.cfg, "moinmessage_gpg_homedir")

def get_signing_users(request):

    "Return a dictionary mapping usernames to signing keys."

    return getWikiDict(
        getattr(request.cfg, "moinmessage_gpg_signing_users_page", "MoinMessageSigningUserDict"),
        request)

def get_relays(request):

    "Return a dictionary mapping relays to URLs."

    return getWikiDict(
        getattr(request.cfg, "moinmessage_gpg_relays_page", "MoinMessageRelayDict"),
        request)

def get_relaying_user(request):

    "Return the relaying username or None if no relaying user is configured."

    return getattr(request.cfg, "moinmessage_gpg_relaying_user")

def get_recipients(request, main=False, sending=True, fetching=True):

    """
    Return the recipients dictionary by first obtaining the page in which it
    is stored. This page may either be a subpage of the user's home page, if
    stored on this wiki, or it may be relative to the site root.

    When 'main' is specified and set to a true value, only a dictionary under
    the site root is consulted.

    When 'sending' or 'fetching' is specified and set to a false value, any
    recipients of the indicated type will be excluded from the result of this
    function.

    The name of the subpage is defined by the configuration setting
    'moinmessage_gpg_recipients_page', which if absent is set to
    "MoinMessageRecipientsDict".
    """

    subpage = getattr(request.cfg, "moinmessage_gpg_recipients_page", "MoinMessageRecipientsDict")

    if not main:
        homepage = get_local_homepage(request, None)
        recipients = getWikiDict("%s/%s" % (homepage, subpage), request)
        if recipients:
            return filter_recipients(recipients, sending, fetching)

    return filter_recipients(getWikiDict(subpage, request), sending, fetching)

def get_local_homepage(request, username):

    """
    Using the 'request', return the homepage for the given 'username' or None if
    no local homepage exists for the user. If 'username' is None, the current
    user's homepage is located.
    """

    homedetails = getInterwikiHomePage(request, username)

    if homedetails:
        homewiki, homepage = homedetails
        if homewiki == "Self":
            return homepage

    return None

def get_username_for_fingerprint(request, fingerprint):

    """
    Using the 'request', return the username corresponding to the given key
    'fingerprint' or None if no correspondence is present in the mapping page.
    """

    # Since this function must be able to work before any user has been
    # identified, the wikidict operation uses superuser privileges.

    gpg_users = getWikiDict(
        getattr(request.cfg, "moinmessage_gpg_users_page", "MoinMessageUserDict"),
        request,
        superuser=True
        )

    if gpg_users and gpg_users.has_key(fingerprint):
        return gpg_users[fingerprint]
    else:
        return None

def get_update_actions_for_user(request):

    """
    For the user associated with the 'request', return the permitted actions for
    the user in the form of a list of "action:pagename" identifiers.
    """

    if not request.user or not request.user.valid:
        return []

    actions = getWikiDict(
        getattr(request.cfg, "moinmessage_user_actions_page", "MoinMessageUserActionsDict"),
        request
        )

    username = request.user.name

    if actions and actions.has_key(username):
        return getIdentifiers(actions[username])
    else:
        return []

def get_recipient_details(request, recipient, main=False, fetching=False):

    """
    Using the 'request', return a dictionary of details for the specified
    'recipient'. If no details exist, raise a MoinMessageRecipientError
    exception.

    When 'main' is specified and set to a true value, only the recipients
    dictionary under the site root is consulted.

    When 'fetching' is specified and set to a true value, the recipient need
    not have a "type" or "location" defined, but it must have a "fingerprint"
    defined.
    """

    _ = request.getText

    recipients = get_recipients(request, main)
    if not recipients:
        raise MoinMessageNoRecipients, _("No recipients page is defined for MoinMessage.")

    recipient_details = recipients.get(recipient)
    if not recipient_details:
        raise MoinMessageUnknownRecipient, _("The specified recipient is not present in the list of known contacts.")

    parameters = parseDictEntry(recipient_details, RECIPIENT_PARAMETERS)

    type = parameters.get("type")
    location = parameters.get("location")
    fingerprint = parameters.get("fingerprint")

    if type in (None, "none") and not fetching:
        raise MoinMessageBadRecipient, _("The recipient details are missing a destination type.")

    if type not in (None, "none") and not location:
        raise MoinMessageBadRecipient, _("The recipient details are missing a location for sent messages.")

    if type in ("url", "relay", None, "none") and not fingerprint:
        raise MoinMessageBadRecipient, _("The recipient details are missing a fingerprint for sending messages.")

    return parameters

def filter_recipients(recipients, sending, fetching):

    """
    Return a copy of the given 'recipients' dictionary retaining all entries
    that apply to the given 'sending' and 'fetching' criteria.
    """

    result = {}
    for recipient, details in recipients.items():
        parameters = parseDictEntry(details, RECIPIENT_PARAMETERS)

        if not fetching and parameters.get("type") in (None, "none"):
            continue
        if not sending and not parameters.get("fingerprint"):
            continue

        result[recipient] = details

    return result

# Moin-specific messaging functions.

def get_signing_identity(request, type):

    """
    Using the 'request' and indicated delivery 'type', return the appropriate
    signing identity for a message, or None if no additional signature is to be
    performed.
    """

    signing_users = get_signing_users(request)
    signer = signing_users and signing_users.get(request.user.name)

    # Send relayed messages with an extra signature.

    if type == "relay":
        relaying_user = get_relaying_user(request)

        # Signing with the current identity if no special relaying user is
        # defined.

        if relaying_user and request.user.name != relaying_user:
            signer = signing_users and signing_users.get(relaying_user)

        return signer

    return None

# Access to static Moin content.

htdocs = None

def get_htdocs(request):

    "Use the 'request' to find the htdocs directory."

    global htdocs
    htdocs = getStaticContentDirectory(request)

    if not htdocs:
        htdocs_in_cfg = getattr(request.cfg, "moinmessage_static_files", None)
        if htdocs_in_cfg and exists(htdocs_in_cfg):
            htdocs = htdocs_in_cfg
            return htdocs

    return htdocs

# Special message formatters.

def unquoteWikinameURL(url, charset=config.charset):

    """
    The inverse of wikiutil.quoteWikinameURL, returning the page name referenced
    by the given 'url', with the page name assumed to be encoded using the given
    'charset' (or default charset if omitted).
    """

    return unicode(urllib.unquote(url), encoding=charset)

def getAttachmentFromURL(url, request):

    """
    Return a (full path, attachment filename) tuple for the attachment
    referenced by the given 'url', using the 'request' to interpret the
    structure of 'url'.

    If 'url' does not refer to an attachment on this wiki, None is returned.
    """

    # Detect static resources.

    htdocs_dir = get_htdocs(request)

    if htdocs_dir:
        prefix = request.cfg.url_prefix_static

        # Normalise the 

        if not prefix.endswith("/"):
            prefix += "/"

        if url.startswith(prefix):
            filename = url[len(prefix):]

            # Obtain the resource path.

            path = abspath(join(htdocs_dir, filename))

            if exists(path):
                return path, taintfilename(filename)

    # Detect attachments and other resources.

    script = request.getScriptname()

    # Normalise the URL.

    if not script.endswith("/"):
        script += "/"

    # Reject URLs outside the wiki.

    if not url.startswith(script):
        return None

    path = url[len(script):].lstrip("/")
    try:
        qpagename, qs = path.split("?", 1)
    except ValueError:
        qpagename = path
        qs = None

    pagename = unquoteWikinameURL(qpagename)
    qs = qs and parseQueryString(qs) or {}

    filename = qs.get("target") or qs.get("drawing")
    filename = taintfilename(filename)

    # Obtain the attachment path.

    path = AttachFile.getFilename(request, pagename, filename)
    return path, filename

class OutgoingHTMLFormatter(text_html.Formatter):

    """
    Handle outgoing HTML content by identifying attachments and rewriting their
    locations. References to bundled attachments are done using RFC 2111:

    https://tools.ietf.org/html/rfc2111

    Messages employing references between parts are meant to comply with RFC
    2387:

    https://tools.ietf.org/html/rfc2387
    """

    def __init__(self, request, **kw):
        text_html.Formatter.__init__(self, request, **kw)
        self.attachments = []

    def add_attachment(self, location):
        details = getAttachmentFromURL(location, self.request)
        if details:
            pos = len(self.attachments)
            self.attachments.append(details)
            return "cid:attachment%d" % pos
        else:
            return None

    def image(self, src=None, **kw):
        src = src or kw.get("src")
        ref = src and self.add_attachment(src)
        return text_html.Formatter.image(self, ref or src, **kw)

    def transclusion(self, on, **kw):
        if on:
            data = kw.get("data")
            kw["data"] = data and self.add_attachment(data)
        return text_html.Formatter.transclusion(self, on, **kw)

# vim: tabstop=4 expandtab shiftwidth=4
